# Imports
import webiopi
from datetime import datetime
import time
import sys
import os
sys.path.append(os.path.dirname(__file__))
from settings import PiColourSettings, Timing, Location, HsvColour, RgbColour
import json

# Enable debug output
#webiopi.setDebug()

gpio = webiopi.GPIO
pca = webiopi.deviceInstance("pwm0")
settings = None
notOE = 4 # NOT Output Enable pin

loopPeriod = 0.01
lastLoopTime = datetime.now()
enableCheckPeriod = 1.0
lastEnableCheckTime = datetime.now()
cachedEnable = False
savePeriod = 60.0
lastSaveTime = datetime.now()
fadeValue = 0.0

# Called by WebIOPi at script loading
def setup():
	webiopi.debug("Script setup")
	global settings
	configfile = os.path.join(os.path.dirname(__file__), 'config')
	settings = PiColourSettings(configfile)

	for channel in settings.channels:
		writeColour(channel)

	setOutputEnable(True)

# Looped by WebIOPi
def loop():
	global lastLoopTime
	global lastEnableCheckTime
	global lastSaveTime
	global cachedEnable
	global fadeValue

	now = datetime.now()
	loopDelta = (now - lastLoopTime).total_seconds()
	enableCheckDelta = (now - lastEnableCheckTime).total_seconds()
	saveDelta = (now - lastSaveTime).total_seconds()
	lastLoopTime = now

	if enableCheckDelta >= enableCheckPeriod:
		lastEnableCheckTime = now
		cachedEnable = shouldEnable()

	if cachedEnable:
		if fadeValue < 1.0:
			if fadeValue == 0.0:
				setOutputEnable(True)

			fadeValue = min(fadeValue + (loopDelta / settings.fadeDuration), 1.0)
			writeAllColours()
	else:
		if fadeValue > 0.0:
			fadeValue = max(fadeValue - (loopDelta / settings.fadeDuration), 0.0)
			writeAllColours()

			if fadeValue == 0.0:
				setOutputEnable(False)

	if settings.updated and saveDelta >= savePeriod:
		lastSaveTime = now
		settings.writeConfig()

	webiopi.sleep(loopPeriod)

# Called by WebIOPi at server shutdown
def destroy():
	if settings.updated:
		settings.writeConfig()

def setOutputEnable(enabled):
	#TODO: Should fade in/out instead of simply turning on/off immediately.
	gpio.digitalWrite(notOE, gpio.LOW if enabled else gpio.HIGH)

def getOutputEnable():
	return not gpio.digitalRead(notOE)

def shouldEnable():
	if settings.override == 'enabled':
		return True
	elif settings.override == 'disabled':
		return False

	s = settings.location.sun
	now = datetime.now()
	onTimes = settings.onTime.toTimes(s, now)
	offTimes = settings.offTime.toTimes(s, now)

	#If the onTime to check is the latest one that has passed...
	if now >= onTimes['tomorrow']:
		onTimeTest = onTimes['tomorrow']
	elif now >= onTimes['today']:
		onTimeTest = onTimes['today']
	else:
		onTimeTest = onTimes['yesterday']

	#... then the offTime is the earliest one that is after the onTime.
	if offTimes['yesterday'] >= onTimeTest:
		offTimeTest = offTimes['yesterday']
	elif offTimes['today'] >= onTimeTest:
		offTimeTest = offTimes['today']
	else:
		offTimeTest = offTimes['tomorrow']

	return now >= onTimeTest and now < offTimeTest

@webiopi.macro('GET', 'override', 'application/json')
def getOverride():
	return json.dumps(settings.override)

@webiopi.macro('PUT', 'override')
def setOverride(data):
	settings.override = data.lower()
	settings.updated = True

@webiopi.macro('GET', 'fadeDuration', 'application/json')
def getFadeDuration():
	return json.dumps(settings.fadeDuration)

@webiopi.macro('PUT', 'fadeDuration')
def setFadeDuration(data):
	settings.fadeDuration = data
	settings.updated = True

@webiopi.macro('GET', 'onTime', 'application/json')
def getOnTime():
	return settings.onTime.toJson()

@webiopi.macro('PUT', 'onTime')
def setOnTime(type, offset, time):
	settings.onTime.update({'type': type, 'offset': offset, 'time': time})
	settings.updated = True

@webiopi.macro('GET', 'offTime', 'application/json')
def getOffTime():
	return settings.offTime.toJson()

@webiopi.macro('PUT', 'offTime')
def setOffTime(type, offset, time):
	settings.offTime.update({'type': type, 'offset': offset, 'time': time})
	settings.updated = True

@webiopi.macro('GET', 'location', 'application/json')
def getLocation():
	return settings.location.toJson()

@webiopi.macro('PUT', 'location')
def setLocation(latitude, longitude):
	settings.location.update({'latitude': latitude, 'longitude': longitude})
	settings.updated = True

@webiopi.macro('GET', 'colour', 'application/json')
def getColour(path):
	channelIndex = int(path)
	channel = settings.channels[channelIndex]

	return json.dumps(channel.colour.toDict())

@webiopi.macro('PUT', 'colour')
def setColour(path, h, s, v):
	channelIndex = int(path)
	channel = settings.channels[channelIndex]
	channel.colour.update({ 'h': h, 's': s, 'v': v })
	settings.updated = True

	writeColour(channel)

def writeColour(channel):
	adjustedColour = channel.getAdjustedRgb(fadeValue)

	for key in adjustedColour:
		pca.pwmWriteFloat(channel.pwmChannels[key], adjustedColour[key])

def writeAllColours():
	for channel in settings.channels:
		writeColour(channel)

@webiopi.macro('GET', 'channels', 'application/json')
def getChannels():
	return json.dumps(list(map(lambda channel: channel.toDict(), settings.channels)))

@webiopi.macro('GET', 'tint', 'application/json')
def getTint(path):
	channelIndex = int(path)
	channel = settings.channels[channelIndex]

	return json.dumps(channel.tint.toDict())

@webiopi.macro('PUT', 'tint')
def setTint(path, r, g, b):
	channelIndex = int(path)
	channel = settings.channels[channelIndex]
	channel.tint.update({ 'r': r, 'g': g, 'b': b })
	settings.updated = True

	writeColour(channel)

