from configparser import ConfigParser
from datetime import datetime, timedelta
import json
from collections import OrderedDict
from abc import ABCMeta, abstractmethod
import math
from sunrise import sun

class PiColourSettings:
	def __init__(self, filename):
		self.filename = filename
		self.updated = False
		self.readConfig();

	def readConfig(self):
		config = self.__getConfig()

		self.override = self.__readOption(config, 'OUTPUT', 'override')
		self.fadeDuration = self.__readOption(config, 'OUTPUT', 'fadeDuration')
		self.location = self.__readSection(config, 'LOCATION', Location)
		self.onTime = self.__readSection(config, 'ONTIME', Timing)
		self.offTime = self.__readSection(config, 'OFFTIME', Timing)
		self.channels = self.__readSectionArray(config, 'CHANNEL', ChannelSettings)

		self.updated = False

	def writeConfig(self):
		config = self.__getConfig()

		self.__writeOption(config, 'OUTPUT', 'override', self.override)
		self.__writeOption(config, 'OUTPUT', 'fadeDuration', self.fadeDuration)
		self.__writeSection(config, 'LOCATION', self.location)
		self.__writeSection(config, 'ONTIME', self.onTime)
		self.__writeSection(config, 'OFFTIME', self.offTime)
		self.__writeSectionArray(config, 'CHANNEL', self.channels)

		file = open(self.filename, 'w')
		config.write(file)
		file.close()
		self.updated = False

	def __getConfig(self):
		config = ConfigParser()
		config.optionxform = str
		config.read(self.filename)

		return config

	def __readSection(self, config, section, type):
		dict = {}

		for option in config.options(section):
			dict[option] = self.__readOption(config, section, option)

		return type(dict)

	def __readOption(self, config, section, option):
		return json.loads(config.get(section, option), object_pairs_hook=OrderedDict)

	def __writeSection(self, config, section, values):
		dict = values.toDict()

		for option, value in dict.items():
			self.__writeOption(config, section, option, value)

	def __writeOption(self, config, section, option, value):
		config.set(section, option, json.dumps(value))

	def __readSectionArray(self, config, sectionPrefix, type):
		sections = [section for section in config.sections() if section.startswith(sectionPrefix)]
		values = [None] * len(sections)

		for section in sections:
			index = int(section.split('_')[1])
			values[index] = self.__readSection(config, section, type)

		return values

	def __writeSectionArray(self, config, sectionPrefix, values):
		for index, value in enumerate(values):
			section = sectionPrefix + '_' + str(index)
			self.__writeSection(config, section, value)

class Setting:
	__metaclass__ = ABCMeta

	def __init__(self, dict):
		self.update(dict)

	def update(self, dict):
		for key, value in dict.items():
			setattr(self, key, value)

	@abstractmethod
	def toDict(self):
		pass

	def toJson(self):
		return json.dumps(self.toDict())

class HsvColour(Setting):
	def toDict(self):
		dict = OrderedDict()
		dict['h'] = self.h
		dict['s'] = self.s
		dict['v'] = self.v

		return dict

	def __eq__(self, other):
		return self.h == other['h'] and self.s == other['s'] and self.v == other['v']

	def __ne__(self, other):
		return self.h != other['h'] or self.s != other['s'] or self.v != other['v']

	def toRgb(self):
		rgb = { 'r': 0.0, 'g': 0.0, 'b': 0.0}
		Chroma = self.s * self.v
		Hdash = self.h / (math.pi / 3.0)
		X = Chroma * (1.0 - math.fabs((Hdash % 2.0) - 1.0))

		if Hdash < 1.0:
			rgb['r'] = Chroma
			rgb['g'] = X
		elif Hdash < 2.0:
			rgb['r'] = X
			rgb['g'] = Chroma
		elif Hdash < 3.0:
			rgb['g'] = Chroma
			rgb['b'] = X
		elif Hdash < 4.0:
			rgb['g']= X
			rgb['b'] = Chroma
		elif Hdash < 5.0:
			rgb['r'] = X
			rgb['b'] = Chroma
		elif Hdash < 6.0:
			rgb['r'] = Chroma
			rgb['b'] = X

		Min = self.v - Chroma

		rgb['r'] += Min
		rgb['g'] += Min
		rgb['b'] += Min

		return rgb

class RgbColour(Setting):
	def toDict(self):
		dict = OrderedDict()
		dict['r'] = self.r
		dict['g'] = self.g
		dict['b'] = self.b

		return dict

	def __eq__(self, other):
		return self.r == other['r'] and self.g == other['g'] and self.b == other['b']

	def __ne__(self, other):
		return self.r != other['r'] or self.g != other['g'] or self.b != other['b']

	def toHsv(self):
		hsv = { 'h': 0.0, 's': 0.0, 'v': 0.0 }
		Min = min(self.r, self.g, self.b)
		Max = max(self.r, self.g, self.b)
		Chroma = Max - Min

		#If Chroma is 0, then S is 0 by definition, and H is undefined but 0 by convention.
		if Chroma != 0:
			if self.r == Max:
				hsv['h'] = (self.g - self.b) / Chroma

				if hsv['h'] < 0.0:
					hsv['h'] += 6.0
			elif self.g == Max:
				hsv['h'] = ((self.b - self.r) / Chroma) + 2.0
			else: #self.b == Max
				hsv['h'] = ((self.r - self.g) / Chroma) + 4.0

			hsv['h'] *= (math.pi / 3.0)
			hsv['s'] = Chroma / Max

		hsv['v'] = Max

		return hsv

class Location(Setting):
	def update(self, dict):
		super(Location, self).update(dict)
		self.sun = sun(lat=self.latitude,long=self.longitude)

	def toDict(self):
		dict = OrderedDict()
		dict['latitude'] = self.latitude
		dict['longitude'] = self.longitude

		return dict

class Timing(Setting):
	def __init__(self, dict):
		self.__timeFormat = '%H:%M'
		self.update(dict)

	def update(self, dict):
		self.type = dict['type']
		self.offset = timedelta(minutes=dict['offset'])
		self.time = datetime.strptime(dict['time'], self.__timeFormat).time()

	def toDict(self):
		dict = OrderedDict()
		dict['type'] = self.type
		dict['offset'] = self.offset.total_seconds() / 60.0
		dict['time'] = self.time.strftime(self.__timeFormat)

		return dict

	def toTimes(self, sun, now):
		time = now.time()
		today = now.date()
		yesterday = today - timedelta(days=1)
		tomorrow = today + timedelta(days=1)

		if self.type == 'sunrise':
			sunriseTimeYesterday = sun.sunrise(when=datetime.combine(yesterday, time))
			sunriseTimeToday = sun.sunrise(when=datetime.combine(today, time))
			sunriseTimeTomorrow = sun.sunrise(when=datetime.combine(tomorrow, time))

			timeYesterday = datetime.combine(yesterday, sunriseTimeYesterday) + self.offset
			timeToday = datetime.combine(today, sunriseTimeToday) + self.offset
			timeTomorrow = datetime.combine(tomorrow, sunriseTimeTomorrow) + self.offset
		elif self.type == 'sunset':
			sunsetTimeYesterday = sun.sunset(when=datetime.combine(yesterday, time))
			sunsetTimeToday = sun.sunset(when=datetime.combine(today, time))
			sunsetTimeTomorrow = sun.sunset(when=datetime.combine(tomorrow, time))

			timeYesterday = datetime.combine(yesterday, sunsetTimeYesterday) + self.offset
			timeToday = datetime.combine(today, sunsetTimeToday) + self.offset
			timeTomorrow = datetime.combine(tomorrow, sunsetTimeTomorrow) + self.offset
		elif self.type == 'fixed':
			timeYesterday = datetime.combine(yesterday, self.time)
			timeToday = datetime.combine(today, self.time)
			timeTomorrow = datetime.combine(tomorrow, self.time)
		else:
			raise ValueError('Wrong Timing type: ' + str(self.type))

		return {'yesterday': timeYesterday, 'today': timeToday, 'tomorrow': timeTomorrow}

class ChannelSettings(Setting):
	def __init__(self, dict):
		self.update(dict)
		self.tint = RgbColour(dict['tint'])
		self.colour = HsvColour(dict['colour'])

	def toDict(self):
		dict = OrderedDict()
		dict['name'] = self.name
		dict['description'] = self.description
		dict['pwmChannels'] = self.pwmChannels
		dict['common'] = self.common
		dict['gamma'] = self.gamma
		dict['tint'] = self.tint.toDict()
		dict['colour'] = self.colour.toDict()

		return dict

	#Returns the colour adjusted by gamma and tint, so it will look correct when sent to the LEDs.
	def getAdjustedRgb(self, fade):
		rgbColour = self.colour.toRgb()
		tint = self.tint.toDict()
		adjustedColour = {}

		for key in rgbColour:
			adjustedColour[key] = self.__adjustRgbValue(rgbColour[key] * fade, tint[key], self.gamma, self.common)

		return adjustedColour

	@staticmethod
	def __adjustRgbValue(value, tint, gamma, common):
		raw = (1.0 - value) if (common == 'anode') else value;

		return (raw ** gamma) * tint;
